/*
 * Copyright (C) 2000 Ken McCrary
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. (www.gnu.org)
 */
package com.kenmccrary.jtella;

import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;

import com.kenmccrary.jtella.util.Log;

/**
 *  Represents a connection to a servant on the network
 *
 *
 */
// TODO rename this class
class BaseConnection implements Runnable
{
  private static String greeting = "GNUTELLA CONNECT/0.4\n\n";
  private static String SERVER_READY = "GNUTELLA OK";

  private boolean shutdownFlag = false;
  private Socket socket;
  private DataInputStream inputStream;
  private DataOutputStream outputStream;
  private ConnectionControl connectionControl;
  private String host;

  /**
   *  Construct the Connection;
   *
   *  @param connectionControl component responsible for controlling connections
   *  @param host can be a machine name or IP address
   *  @param port port to use
   */
  public BaseConnection(ConnectionControl connectionControl,
                        String host, 
                        int port) throws UnknownHostException,
                                                      IOException
  {
    this.connectionControl = connectionControl;
    this.host = host;
    socket = new Socket(InetAddress.getByName(host),
                        port);
    socket.setSoTimeout(15000);
    socket.setTcpNoDelay(true);

    inputStream = new DataInputStream(socket.getInputStream());
    outputStream = new DataOutputStream(socket.getOutputStream());

  }

  /**
   *  Retrieve the host this connection links to
   *
   */
  String getHost()
  {
    return host;
  }

  /**
   *  Stops the connection and cleans up
   *
   */
  void shutdown()
  {
    shutdownFlag = true;
    try
    {
      if ( null != outputStream ) 
      {
        outputStream.close();
      }
      if ( null != inputStream ) 
      {
        inputStream.close();
      }
      if ( null != socket ) 
      {
        socket.close();
      }

    }
    catch (Exception e)
    {

    }
  }

  /** 
   *  Starts a connection to a node on the network,
   *  does initial handshaking
   *
   *  @return true for a good start, false for failure
   */
  public boolean start()
  {
    boolean result = true;

    try
    {
      Log.getLog().logDebug("Sending greeting to : " + host);
      byte[] greetingData = greeting.getBytes();
      outputStream.write(greetingData, 0, greetingData.length);
      outputStream.flush();

      String response = inputStream.readLine();

      if ( null == response ||
           -1 == response.indexOf(SERVER_READY) )
      {
        Log.getLog().logWarning("Connection rejection: " + host); 
        return false;
      }

      // Pull the second newline from the stream
      int newline = inputStream.readUnsignedByte();

      // Now start monitoring network messages
      Thread thread = new Thread(this, "Connection");
      thread.start();

    }
    catch (Exception e)
    {
      result = false;
      e.printStackTrace();
    }
    // todo finally if start false, close connection

    return result;
  }

  /**
   *  Sends a Message through this connection
   *
   *  @param m message
   */
  public void send(Message m) throws IOException
  {
    if ( shutdownFlag ) 
    {
      return;
    }

    Log.getLog().logDebug("Connection sending: " + 
                          m.getType() +
                          " to " +
                          socket.getInetAddress().getHostAddress());

    Log.getLog().logDebug("Sending message: \n" + m.toString());

    m.setTTL((byte) (m.getTTL() - 1));
    byte[] messageBytes = m.getByteArray();

    outputStream.write(messageBytes, 0, messageBytes.length);
    outputStream.flush();
  }

  /**
   *  Sends a message down the connection and sends any response
   *  to <code>MessageReceiver</code>
   *
   */
  public void sendAndReceive(Message m,
                             MessageReceiver receiver) throws IOException
  {
    if ( shutdownFlag ) 
    {
      return;
    }

    Log.getLog().logDebug("Connection sendAndReceive: " + 
                           m.getType() +
                           " to " +
                           socket.getInetAddress().getHostAddress());

    ConnectionControl.getRouter().routeBack(m, receiver);

    send(m);
  }


  /**
   *  Connection monitoring thread
   *
   *
   */
  public void run()
  {

    try
    {
      while( !shutdownFlag )
      {
        // Read a message 
        byte[] message = new byte[Message.SIZE];
        for (int i = 0; i < message.length; i++) 
        {
          message[i] = (byte)inputStream.readUnsignedByte();
        }

        Message readMessage = MessageFactory.createMessage(message);

        if ( null == readMessage ) 
        {
          Log.getLog().logError("MessageFactory.createMessage() returned null");
          continue;
        }

        int payloadSize = readMessage.getPayloadLength();

        if ( !readMessage.validatePayloadSize() ) 
        {
          handleConnectionError();
          Log.getLog().logInformation("Received invalid message from: " + 
                                      host +
                                      ", message type: " +
                                      readMessage.getType());
          continue;
        }

        if (payloadSize > 0 ) 
        {
          short[] payload = new short[payloadSize];
          // Read the payload
          for (int p = 0; p < payloadSize; p++ ) 
          {
            payload[p] = (short)inputStream.readUnsignedByte();
          }

          readMessage.addPayload(payload);
        }
        
        Log.getLog().logDebug("Read message from " + 
                              host +
                              " : " +
                              readMessage.toString());

        // Message is read, route it
        ConnectionControl.getRouter().route(readMessage, this);

      }
      
    }
    catch (Exception e)
    {
      handleConnectionError();
    }
  }


  /**
   *  Handles a serious error on the connection
   *
   */
  void handleConnectionError()
  {
    Log.getLog().logInformation("Shutting down connection: " + host);
    connectionControl.reportFailedConnection(this);
    shutdown();
  }
}
 
 
