/*
 * Copyright (C) 2000 Ken McCrary
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. (www.gnu.org)
 */
package com.kenmccrary.jtella;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;
import java.util.Enumeration;

import com.kenmccrary.jtella.util.Log;

/**
 *  Responsible for maintaining connection(s) to the
 *  network
 *
 */
class ConnectionControl extends Thread
{
  private HostCache hostCache;
  private boolean shutdownFlag;
  private LinkedList connectionList;
  private int desiredConnections;
  private static Router router;
  private int connectionIndex;
  private Vector connectionListeners;

  ConnectionControl(HostCache hostCache, int desiredConnections)
  {
    super("Connection Control Thread");
    this.hostCache = hostCache;
    connectionList = new LinkedList();
    router = new Router(connectionList, hostCache);
    router.start();
    this.desiredConnections = desiredConnections;
    connectionIndex = -1;
    connectionListeners = new Vector();

  }

  /**
   *  Get the router for handling network messages
   *
   */
  static Router getRouter()
  {
    return router;
  }

  /**
   *  Provides a connection to the network
   *
   */
  BaseConnection getNextConnection()
  {
    if (!hasConnections()) 
    {
      throw new IllegalStateException("No connections");

    }

    connectionIndex++;

    if (connectionIndex >= connectionList.size()) 
    {
      connectionIndex = 0;
    }

    return (BaseConnection)connectionList.get(connectionIndex);
  }

  /**
   *  Check if we are connected to the network
   *
   */
  boolean hasConnections()
  {
    return connectionList.size() > 0 ? true : false;
  }

  /**
   *  Adds a listener interested in connection information
   *
   *  @param listener connection listener
   */
  void addConnectionListener(ConnectionListener listener)
  {
    connectionListeners.addElement(listener);
  }

  /**
   *  Removes a listener interested in connection information
   *
   *  @param listener connection listener
   */
  void removeConnectionListener(ConnectionListener listener)
  {
    connectionListeners.removeElement(listener);
  }

  /**
   *  Invokes the connection listeners with the new connection information
   *
   *  @param event connection event
   */
  void fireConnectionEvent(ConnectionEvent event)
  {
    Vector listenerCopy = (Vector)connectionListeners.clone();
    
    Enumeration e = listenerCopy.elements();

    while ( e.hasMoreElements() ) 
    {
      ConnectionListener listener = (ConnectionListener)e.nextElement();

      listener.connectionChange(event);
    }
  }

  /**
   *  Used to report when a connection fails, response is to create another connection
   *
   */
  void reportFailedConnection(BaseConnection connection)
  {
    Log.getLog().logInformation("Connection failed:" + connection.getHost());
    connectionList.remove(connection);


    ConnectionInformation info = new ConnectionInformation();
    info.setCount(connectionList.size());
    fireConnectionEvent(new ConnectionEvent(info));
  }

  /**
   *  Attempts to maintain connection to the network
   *
   */
  public void run()
  {

    while (!shutdownFlag ) 
    {
      try
      {
        Thread.currentThread().sleep(2000);
      }
      catch(InterruptedException e)
      {
        e.printStackTrace();
      }

      // precondition
      if ( hostCache.size() == 0 ) 
      {
        Log.getLog().logDebug("no hosts cached");
        continue;
      }

      // TODO check for failed connections
      if ( connectionList.size() < desiredConnections )
 
      {
  
        HostCache.Host host = hostCache.nextHost();

        Log.getLog().logInformation("Connection control attempting a connection: " +
                                    host.getIPAddress() +
                                    ":" +
                                    host.getPort());

        try
        {
          BaseConnection connection = new BaseConnection(this,
                                                         host.getIPAddress(),
                                                         host.getPort());  
          boolean startFlag = connection.start();

          if ( startFlag )  
          {
            connectionList.addFirst(connection);
            Log.getLog().logInformation("New connection established on : " +
                                        host.getIPAddress() +
                                        ", current count: " +
                                        connectionList.size());

            ConnectionInformation info = new ConnectionInformation();
            info.setCount(connectionList.size());

            fireConnectionEvent(new ConnectionEvent(info));
          }
          else
          {
            Log.getLog().logInformation("Connection start failed on: " + host.getIPAddress());
          }

        }
        catch(Exception e)
        {
          Log.getLog().logInformation("Connection start failed on: " + host.getIPAddress());
          System.out.println("Connection start failed on: " + host.getIPAddress());
          e.printStackTrace();
        }

        // Host is in use or failed remove it from the cache
        hostCache.removeHost(host);
      }

    }
  }
}
