/*
 * Copyright (C) 2000 Ken McCrary
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. (www.gnu.org)
 */
package com.kenmccrary.jtella;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import com.kenmccrary.jtella.util.Log;

/**
 *  Abstract base class for GNUTella messages
 *
 */
abstract class Message
{
  public final static int SIZE = 23;

  final static byte PING =  (byte)0x00;
  final static byte PONG =  (byte)0x01;
  final static byte PUSH =  (byte)0x40;
  final static byte QUERY = (byte)0x80;
  final static byte QUERYREPLY = (byte)0x81;

  final static int SIZE_PING_PAYLOAD = 0;
  final static int SIZE_PONG_PAYLOAD = 14;
  final static int SIZE_PUSH_PAYLOAD = 26;
  final static int SIZE_QUERY_PAYLOAD = 257;
  final static int SIZE_QUERYREPLY_PAYLOAD = 65536;

  protected GUID guid;
  protected byte type;
  protected byte ttl;
  protected byte hops;
  protected short[] payload;
  protected int payloadSize; 
  
  /**
   *  Constructs a new message
   *
   *
   *  @param type function type
   */
   Message(int type)
   {
     guid = new GUID();
     this.type = (byte)type;
     ttl = 7;
     hops = 0;
   }

  /**
   *  Query the type of message
   *
   *  @return type
   */
  int getType()
  {
    return type;  
  }

   /**
    *  Constructs a message from data read from network
    *
    *  @param rawMessage bytes
    */
   // TODO convert rawMessage to short[]
   Message(byte[] rawMessage)
   {
     StringBuffer buffer = new StringBuffer();

     for (int i = 0; i < rawMessage.length; i++) 
     {
       buffer.append("[" +
                     Integer.toHexString(rawMessage[i]) +
                     "]");

     }

     Log.getLog().logDebug("Raw Message Bytes: " +
                           buffer.toString());

     // Copy the GUID
     byte[] guidData = new byte[16];
     System.arraycopy(rawMessage, 0, guidData, 0, guidData.length);
     guid = new GUID(guidData);

     // Copy the function identifier
     type = (byte)rawMessage[16];

     // Copy the TTL 
     ttl = (byte)rawMessage[17];

     // Copy the hop count
     hops = (byte)rawMessage[18];

     // Copy the payoad size (little endian)
     int byte1 = rawMessage[19];
     int byte2 = rawMessage[20];
     int byte3 = rawMessage[21];
     int byte4 = rawMessage[22];


     payloadSize += byte1;
     payloadSize += (byte2 << 8);
     payloadSize += (byte3 << 16);
     payloadSize += (byte4 << 24);

   }


   /**
    *  Query the GUID 
    *
    *  @return 16 byte array
    */
   
   GUID getGUID()
   {
     return guid;
   }

   /**
    *  Get the Time to live for the message
    *
    */
  int getTTL()
  {
    return ttl;   
  }

  /**
   *  Set the ttl value for the message
   *
   */
  void setTTL(byte ttl)
  {
    this.ttl = ttl;
  }

  /**
   *  Get the hop count for this message
   *
   */
  int getHops()
  {
    return hops;
  }

  /**
   *  Set the hop count for this message, seven is the recommended maximum
   *
   */
  void setHops(byte hops)
  {
    this.hops = hops;
  }
   

   /**
    *   Add a payload to the message
    *
    */
   void addPayload(short[] payload)
   {
     this.payload = payload;
     payloadSize = payload.length;
   }

   /**
    *  Query the payload size for this message
    *
    */
   int getPayloadLength()
   {
     return payloadSize;
   }

   /**
    *  Produces a byte[] suitable for 
    *  GNUTELLA network
    *
    */
   public byte[] getByteArray()
   {
     // TODO handle expection avoid null pointer
     ByteArrayOutputStream byteStream = null;

     try
     {
     
       byteStream = new ByteArrayOutputStream();
       DataOutputStream payloadStream = new DataOutputStream(byteStream);

       // Write the guid
       payloadStream.write(guid.getData());

       // Write the function (payload descriptor)
       payloadStream.writeByte(type);

       // Write the time to live
       payloadStream.writeByte(ttl);

       // Write the hop count
       payloadStream.writeByte(0);

       // Write the Payload size in little-endian
       payloadStream.writeByte(payloadSize);
       payloadStream.writeByte(0);
       payloadStream.writeByte(0);
       payloadStream.writeByte(0);

       // Write the payload
       if ( null != payload ) 
       {
         // Message may not have a payload (ex. Ping)
         for (int i = 0; i < payload.length; i++) 
         {
           byte payloadByte = (byte)payload[i];
           payloadStream.writeByte(payloadByte);
         }
       }
       

       // all done
       payloadStream.close();

     }
     catch (IOException io)
     {
       Log.getLog().log(io);
     }

     return byteStream.toByteArray();
   }
   
   /**
    *  Checks validity of a payloads size
    *
    *  @param message to test
    *  @return true if valid, false otherwise
    */
   boolean validatePayloadSize()
   {
     boolean result = false;
 
     switch ( type ) 
     {
       case PING:
       {
         if ( SIZE_PING_PAYLOAD == payloadSize ) 
         {
           result = true;
         }
         break;
       }

       case PONG:
       {
         if ( SIZE_PONG_PAYLOAD == payloadSize ) 
         {
           result = true;
         }
         break;
       }

       case PUSH:
       {
         if ( SIZE_PUSH_PAYLOAD == payloadSize ) 
         {
           result = true;
         }
         break;
       }

       case QUERY:
       {
         if ( payloadSize < SIZE_QUERY_PAYLOAD ) 
         {
           result = true;
         }
         break;
       }

       case QUERYREPLY:
       {
         if ( payloadSize < SIZE_QUERYREPLY_PAYLOAD) 
         {
           result = true;
         }
         break;
       }
     }
 
     return result;
   }

   /**
    *  String representation of the message
    *
    */
   public String toString()
   {
     StringBuffer message = new StringBuffer();

     message.append("GUID: ");

     byte[] guidData = getGUID().getData();
     for (int i = 0; i < guidData.length; i++) 
     {
       message.append( "[" +
                       Integer.toHexString( guidData[i]) +
                       "]");
     }

     message.append("\n");

     switch (type) 
     {
       case PING:
       {
         message.append("PING message\n");
         break;
       }

       case PONG:
       {
         message.append("PONG message\n");
         break;
       }

       case QUERY:
       {
         message.append("QUERY message\n");
         break;
       }

       case QUERYREPLY:
       {
         message.append("QUERY REPLY message\n");
       }

       default:
       {
         message.append("Unknown message");
         break;
       }
     }

     message.append("Payload: \n");

     if (null != payload) 
     {
       for (int i = 0; i < payload.length; i++) 
       {
         message.append( "[" +
                         Integer.toHexString(payload[i]) + 
                         "]");
       }

       message.append("\n");
     }
     else
     {
       message.append("No payload");
     }


     return message.toString();
   }
}
