/*
 * Copyright (C) 2000 Ken McCrary
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. (www.gnu.org)
 */
package com.kenmccrary.jtella;

import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.IOException;

import java.util.Vector;

import com.kenmccrary.jtella.util.Log;

/**
 *  The NetworkConnection represents a connection to the GNUTella 
 *  <STRONG>network</STRONG>. The connection consists of one or more
 *  socket connections to servants on the network.<p>
 *
 *  The following is an example initiating a NetworkConnection.
 *
 * <PRE>
 *     NetworkConnection c = new NetworkConnection(host, port);
 *     c.addConnectionListener(new ConnectionListenerImpl());
 *     c.start();
 * </PRE>
 */
// TODO rename to network interface
// TODO Merge duplicate code with BaseConnection
public class NetworkConnection implements Runnable
{
  private static String greeting = "GNUTELLA CONNECT/0.4\n\n";
  private static String SERVER_READY = "GNUTELLA OK";

  private String host;
  private int port;
  private Socket socket;
  private DataInputStream inputStream;
  private DataOutputStream outputStream;
  private boolean shutdownFlag;
  private Vector receivedMsgGUIDS;
  private HostCache hostCache;
  private ConnectionControl connectionControl;
  private boolean online;
  private SearchMonitorSession searchMonitorSession;
  
  /**
   *  Construct the Connection;
   *
   *  @param host can be a machine name or IP address
   *  @param port port to use
   */
  public NetworkConnection(String host, int port) throws UnknownHostException,
                                                         IOException
  {
    Log.getLog().logInformation("Network connection initializing");
    Log.getLog().logInformation(System.getProperties().toString());

    receivedMsgGUIDS = new Vector();
    this.host = host;
    this.port = port;
    hostCache = new HostCache();
    online = false;

    socket = new Socket(InetAddress.getByName(host),
                        port);

    inputStream = new DataInputStream(socket.getInputStream());
    outputStream = new DataOutputStream(socket.getOutputStream());

    // TODO configurable count
    connectionControl = new ConnectionControl(hostCache, 2);

  }

  /** 
   *  Starts the connection
   *
   *
   */
  public void start()
  {
    try
    {
      Log.getLog().logDebug("Sending greeting to : " + host);
      byte[] greetingData = greeting.getBytes();
      outputStream.write(greetingData, 0, greetingData.length);
      outputStream.flush();

      String response = inputStream.readLine();

      if ( null == response ||
           -1 == response.indexOf(SERVER_READY) )
      {
        Log.getLog().logWarning("Connection rejection: " + host); 
        // TODO an exception?
        return;
      }

      // read second new line
      int newline = inputStream.readUnsignedByte();

      // TODO check the response

      // Ping, get it started
      PingMessage ping = new PingMessage();
      byte[] bytes = ping.getByteArray();

      outputStream.write(bytes, 0, bytes.length);
      outputStream.flush();

      // Now start monitoring network messages
      Thread thread = new Thread(this, "Connection");
      thread.start();

      connectionControl.start();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  /**
   *  Query if we are online with the network
   *
   *  @return true if online, false otherwise
   */
  public boolean isOnline()
  {
    return connectionControl.hasConnections();
  }

  /**
   *  Adds a listener interested in connection information
   *
   *  @param listener connection listener
   */
  public void addConnectionListener(ConnectionListener listener)
  {
    connectionControl.addConnectionListener(listener);
  }

  /**
   *  Removes a listener interested in connection information
   *
   *  @param listener connection listener
   */
  public void removeConnectionListener(ConnectionListener listener)
  {
    connectionControl.removeConnectionListener(listener);
  }

  /** 
   *  Creates a session to conduct network searches
   *
   *  @param receiver receiver for search responses
   *  @return session
   */
  public SearchSession createSearchSession(MessageReceiver receiver)
  {
    return new SearchSession(connectionControl, receiver);
  }

  /**
   *  Get a search monitor session to monitor query requests
   *  flowing through this network connection, at this time
   *  only one <code>SearchMonitorSession</code> at a time is supported
   *
   *  @param searchReceiver message receiver
   */
  public SearchMonitorSession getSearchMonitorSession(MessageReceiver searchReceiver)
  {

    if ( null == searchMonitorSession ) 
    {
      searchMonitorSession = new SearchMonitorSession(connectionControl, searchReceiver);
    }

    return searchMonitorSession;
  }

  /**
   *  Run the connection thread
   *
   */
  public void run()
  {
    // Send the greeting and begin reading messages
    try
    {

      while( !shutdownFlag )
      {
        // Read a message 

        //if ( 0  == inputStream.available() ) 
        //{
        //  return;
        //}

        byte[] message = new byte[Message.SIZE];
        for (int i = 0; i < message.length; i++) 
        {
          message[i] = (byte)inputStream.readUnsignedByte();
        }

        Message readMessage = MessageFactory.createMessage(message);
        int payloadSize = readMessage.getPayloadLength();

        if (payloadSize > 0 ) 
        {
          short[] payload = new short[payloadSize];
          // Read the payload
          for (int p = 0; p < payloadSize; p++ ) 
          {
            payload[p] = (short)inputStream.readUnsignedByte();
          }

          readMessage.addPayload(payload);
        }
        
        Log.getLog().logDebug("NetWorkConnection Read message from " + 
                              host +
                              " : " +
                              readMessage.toString());

        // Pong messages
        if (readMessage instanceof PongMessage) 
        {
          PongMessage pongMessage = (PongMessage)readMessage;
          hostCache.addHost(pongMessage.getIPAddress(),
                            pongMessage.getPort());
        }

      }
      
    }
    catch (Exception e)
    {
      Log.getLog().log(e);
    }
    finally
    {
      try
      {
        if ( null != socket ) 
        {
          Log.getLog().logInformation("Closing connection to host cache");
          outputStream.close();
          inputStream.close();
          socket.close();
        }
      }
      catch (Exception e)
      {
        Log.getLog().log(e);
      }
    }
  }

}
