/*
 * Copyright (C) 2000 Ken McCrary
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. (www.gnu.org)
 */
package com.kenmccrary.jtella;

import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Stack;

import java.io.IOException;

import com.kenmccrary.jtella.util.Log;

/**
 *  Routes messages read from the network to appropriate
 *  Connections
 *
 */
class Router extends Thread
{
  private static int MAX_ROUTER_TABLE = 5000;
  private static byte MAX_HOPS = (byte)7;
  private static byte MAX_TTL = (byte)50;

  private LinkedList connectionList;
  private HostCache hostCache;
  private RouteTable pingRouteTable;
  private RouteTable queryRouteTable;
  private OriginateTable originateTable;
  private MessageReceiver searchMessageReceiver;
  private Stack messageStack;
  private boolean shutDownFlag;


  /**
   *  Collection of active connections to the network
   *
   */
  Router(LinkedList connectionList, HostCache hostCache)
  {
    this.connectionList = connectionList;
    this.hostCache = hostCache;
    pingRouteTable = new RouteTable(MAX_ROUTER_TABLE);
    queryRouteTable = new RouteTable(MAX_ROUTER_TABLE);
    originateTable = new OriginateTable();
    messageStack = new Stack();
  }

  /**
   *  Stops the operation of the router
   *
   */
  void shutDown()
  {
    shutDownFlag = true;
  }

  /**
   *  Routes a message, used by Connections
   *
   */
  void route(Message m, BaseConnection connection)
  {

    if ( m.getTTL() < 1 ) 
    {
      // expired message
      return;
    }

    RouteMessage message = new RouteMessage(m, connection);

    synchronized (this)
    {
      messageStack.push(message);
      notify();
    }
  }

  /**
   *  Record a message we originate, so we can route it back
   *
   *
   */
  void routeBack(Message m, MessageReceiver receiver)
  {
    originateTable.put(m.getGUID(), receiver);
  }

  /**
   *  Add a listener for all search messages routed, only one of these
   *  are allowed
   *
   */
  void addSearchMessageReceiver(MessageReceiver receiver)
  {
     searchMessageReceiver = receiver;
  }

  /**
   *  Query the next message to route, blocks if no message are available
   *
   *  @return message to route
   */
  RouteMessage getNextMessage()
  {
    synchronized (this)
    {
      while ( messageStack.empty() ) 
      {
        try
        {
          wait();
        }
        catch (InterruptedException ie)
        {
          ie.printStackTrace();
        }
      }

      return (RouteMessage)messageStack.pop();
    }
  }


  /**
   *  Runs along routing messages
   *
   */
  public void run()
  {

    while ( !shutDownFlag ) 
    {
      try
      {
        RouteMessage routeMessage = getNextMessage();

        if ( null == routeMessage ) 
        {
          Log.getLog().logError("Null message in router");
          continue;
        }

        //-----------------------------------------------------------
        // Check if this is a response to a message we generated
        //-----------------------------------------------------------
        if ( originateTable.containsGUID( routeMessage.getMessage().getGUID()) ) 
        {
          Log.getLog().logInformation("Routing response to originated message");

          // Retrieve the message receiver
          Message m = routeMessage.getMessage();

          MessageReceiver receiver = originateTable.get(m.getGUID());

          if ( m instanceof SearchReplyMessage) 
          {
            receiver.receiveSearchReply((SearchReplyMessage)m);
          }
          else
          {
            // TODO PONG REPLY
            Log.getLog().logError("Routeback unknown message");
          }

          continue;
        }

        //-----------------------------------------------------------
        // Don't forward invalid messages
        //-----------------------------------------------------------
        if ( !validateMessage(routeMessage.getMessage())) 
        {
          continue;
        }

        //-----------------------------------------------------------
        // Route the network traffic to our outbound connections
        //-----------------------------------------------------------
        switch ( routeMessage.getMessage().getType() ) 
        {
          case Message.PING:
          {
            Log.getLog().logInformation("Routing ping message");
            routePingMessage(routeMessage);
            break;
          }

          case Message.PONG:
          {
            Log.getLog().logInformation("Routing pong message");
            routePongMessage(routeMessage);
            break;
          }

          case Message.PUSH:
          {
            // TODO
            break;
          }

          case Message.QUERY:
          {
            Log.getLog().logInformation("Routing query message");
            routeQueryMessage(routeMessage);
            break;
          }

          case Message.QUERYREPLY:
          {

            Log.getLog().logInformation("Routing query reply message");
            routeQueryReplyMessage(routeMessage);
            break;
          }
        }
      }
      catch (Exception e)
      {
        // keep running
        Log.getLog().log(e);
      }
    
    }

  }

  // TODO history for routing

  /**
   *  Routes PING messages -
   *  RULES: Route PING messages to all connections
   *  except originator
   *
   */
  void routePingMessage(RouteMessage m)
  {
    // Make a local connection list to avoid concurrency issues
    LinkedList listcopy =  (LinkedList)connectionList.clone();
    ListIterator iterator = listcopy.listIterator(0);

    while ( iterator.hasNext() ) 
    {
      BaseConnection connection = (BaseConnection)iterator.next();

      if ( !connection.equals(m.getConnection()) ) 
      {
        try
        {
          connection.send(m.getMessage());
        }
        catch (IOException io)
        {
          // TODO mechansim to kill this bad connection

          Log.getLog().log(io);
        }
      }
    }

    // History of Pings
    pingRouteTable.put(m.getMessage().getGUID(),
                       m.getConnection());
  }

  /**
   *  Routes PONG messages -
   *  RULES: Route PONG messages only to the connection
   *  the PING arrived on
   *
   */
  void routePongMessage(RouteMessage m)
  {
    BaseConnection originator = pingRouteTable.get(m.getMessage().getGUID());

    if ( null != originator ) 
    {
      try
      {
        originator.send(m.getMessage());

        // Harvest this servant from the pong message
        PongMessage pongMessage = (PongMessage)m.getMessage();
        hostCache.addHost(pongMessage.getIPAddress(), pongMessage.getPort());
      }
      catch (IOException io)
      {
        // TODO update connection control this is a bad message

        Log.getLog().log(io);
      }
    }
    else
    {
      Log.getLog().logError("Null connection in route table");
    }
  }

  /**
   *  Routes QUERY messages - 
   *  RULES: Route QUERY messages to all connections 
   *  except the originator
   *
   */
  void routeQueryMessage(RouteMessage m)
  {
    // Make a local connection list to avoid concurrency issues
    LinkedList listcopy =  (LinkedList)connectionList.clone();
    ListIterator iterator = listcopy.listIterator(0);

    while ( iterator.hasNext() ) 
    {
      BaseConnection connection = (BaseConnection)iterator.next();

      if ( !connection.equals(m.getConnection()) ) 
      {
        try
        {
          connection.send(m.getMessage());
        }
        catch (IOException io)
        {
          // TODO mechansim to kill this bad connection

          Log.getLog().log(io);
        }
      }
    }

    // History of Queries
    queryRouteTable.put(m.getMessage().getGUID(),
                        m.getConnection());

    // inform any search listener
    if ( null != searchMessageReceiver ) 
    {
      searchMessageReceiver.receiveSearch((SearchMessage)m.getMessage());
    }
  }

  /**
   *  Routes QUERY REPLY messages -
   *  Rules: Route Query Replys to the connection
   *  which had the Query
   *
   */
  void routeQueryReplyMessage(RouteMessage m)
  {
    BaseConnection originator = queryRouteTable.get(m.getMessage().getGUID());

    if ( null != originator ) 
    {
      try
      {
        originator.send(m.getMessage());
      }
      catch (IOException io)
      {
        // TODO update connection control this is a bad message

        Log.getLog().log(io);
      }
    }
    else
    {
      Log.getLog().logError("Null connection in route table");
    }
  }

  /** 
   *  Performs some validation against network trafic
   *
   *  @return true if the message is acceptable, false otherwise
   */
  boolean validateMessage(Message m)
  {

    //---------------------------------------------------------------
    //  The idea is to limit trafic, making sure hops doesn't exceed
    //  a reasonable amount - 7
    //---------------------------------------------------------------
    if ( m.getHops() > MAX_HOPS ) 
    {
      Log.getLog().logInformation("Router dropped message exceeding max hops");
      return false;
    }

    if ( m.getTTL() > MAX_TTL) 
    {
      Log.getLog().logInformation("Router dropped message exceeding max ttl");
      return false;
    }


    if ( m.getTTL() > MAX_HOPS &&
         m.getTTL() < MAX_TTL ) 
    {
      Log.getLog().logInformation("Router adjusted message ttl to 7");
      m.setTTL(MAX_HOPS);
    }

    if ( (m.getTTL() + m.getHops()) > MAX_HOPS ) 
    {
      Log.getLog().logInformation("Router adjusted message ttl to 7");
      m.setTTL((byte)(MAX_HOPS - m.getHops()));
    }
    

    return true;
  }

  /**
   *  Records a message to route
   *
   */
  class RouteMessage
  {
    Message m;
    BaseConnection connection;

    RouteMessage(Message m, BaseConnection connection)
    {
      this.m = m;
      this.connection = connection;
    }
    
    /**
     *
     *
     */
    Message getMessage()
    {
      return m;
    }

    /**
     *
     *
     */
    BaseConnection getConnection()
    {
      return connection;
    }
  }
}
